// Author: Pete Kirkham
//
// quad store.
//
// file-reader with SQL escaped quoted string

#ifndef _SQL_READER_H_
#define _SQL_READER_H_

#include "io/delegating_reader.h"
#include "io/reader_base.h"

#include "core/triple_target.h"

template <class R> 
class sql_reader : public delegating_reader<R>
{
private:
    typedef  delegating_reader<R> super;

public:
    sql_reader (R& reader) : delegating_reader<R>(reader) { }
    
    inline size_t   read_quoted_string (char* string_buf)
    {
        int character(super::read());
        int index(0);
        int quote(-1);

        string_buf[0] = 0;

        while (character == ' ')
        {
            character = super::read();
        }

        if ((character == '\'') || (character == '\"'))
        {
            quote = character;
        }
        else
        {
            super::set_failed();

            return 0;
        }

        for (;;)
        {
            character = super::read();
            if (character == -1)
            {
                break;
            }

            if (character == quote)
            {
                character = super::read();

                if (character != quote)
                {
                    break;
                }
            }
            else if (character == '\\')
            {
                character = super::read();

                switch (character)
                {
                    case 'n': character = '\n'; break;
                    case 'r': character = '\r'; break;
                    case 't': character = '\t'; break;
                    default:  character = character;  // no-op;
                }

            }

            if (index < MAX_INDEX)
            {
                string_buf[index++] = character;
            }
        }

        string_buf[index] = 0;

        return index;
    }

    inline char read_quoted_char ()
    {
        int character(super::read());

        while (character == ' ')
        {
            character = super::read();
        }

        if (character == '\'')
        {
            character = super::read();

            if (super::read() == '\'')
            {
                super::read();
                return character;
            }
        }

        super::set_failed();

        return 0;
    }
    
    inline size_t read_int_as_string (char* buf)
    {
        int digit(super::read());
     
        while (digit == ' ')
        {
            digit = super::read();
        }
        
        size_t index(0);

        if ((digit == '-') || ((digit >= '0') && (digit <= '9')))
        {
            buf[index++] = digit;
        }
        else
        {
            index = 0;
        }

        while (((digit = super::read()) >= '0') && (digit <= '9'))
        {
            buf[index++] = digit;
        }

        return index;
    }

    // read sql, write to writer
    template <typename W>
    void pipe_to (W& target, size_t max_read)
    {
        char    subject_buf[BLOCK_SIZE];
        char    predicate_buf[BLOCK_SIZE];
        char    object_buf[BLOCK_SIZE];
        char    graph_buf[BLOCK_SIZE];
        size_t  subject_length;
        size_t  predicate_length;
        size_t  object_length;
        size_t  graph_length;

        size_t  ntriples(0);
        clock_t start_time(clock());

        if (!super::expect("INSERT INTO `triple_all` VALUES "))
        {
            std::cerr << "error searching for table start" << std::endl;
            exit(1);
        }

        size_t  mibs_read_step(5 * MiB);
        size_t  mibs_read(mibs_read_step);

        // read in SQL rows
        for (;;)
        {
            if (!super::expect('('))
            {
                std::cerr << "expected '(' at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                break;
            }

            // `idswd_source` int(11) NOT NULL default '0',
            graph_length = read_int_as_string(graph_buf);

            if (super::failed())
            {
                std::cerr << "expected integer at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                break;
            }

            //~ if (!super::expect(','))
            //~ {
                //~ std::cerr << "expected ',' at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                //~ break;
            //~ }

            // `subject` varchar(250) character set utf8 collate utf8_bin NOT NULL default '',
            subject_length = read_quoted_string(subject_buf);

            if (super::failed())
            {
                std::cerr << "expected string at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                break;
            }

            // `predicate` varchar(250) character set utf8 collate utf8_bin NOT NULL default '',
            predicate_length = read_quoted_string(predicate_buf);

            if (super::failed())
            {
                std::cerr << "expected string at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                break;
            }

            // `object` text NOT NULL,
            object_length = read_quoted_string(object_buf);

            if (super::failed())
            {
                std::cerr << "expected string at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                break;
            }

            // `type_subject` char(1) NOT NULL default 'u',
            char subject_type(read_quoted_char());

            if (super::failed())
            {
                std::cerr << "expected char at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                break;
            }

            // `type_object` char(1) NOT NULL default 'u',
            char object_type(read_quoted_char());

            if (super::failed())
            {
                std::cerr << "expected char at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                break;
            }

            // output triple
            if ((subject_type != 'u') && (subject_type != 'b'))
            {
                std::cerr << "Unexpected subject type: " << subject_type << std::endl;
                break;
            }

            if ((object_type != 'u') && (object_type != 'b') && (object_type != 'l'))
            {
                std::cerr << "Unexpected object type: " << object_type << std::endl;
                break;
            }

            target.add_triple( subject_buf, subject_length,
                               subject_type == 'u' ? triple_target::type_uri : triple_target::type_blank_node,
                               predicate_buf, predicate_length,
                               object_buf, object_length,
                               object_type == 'u' ? triple_target::type_uri : (object_type == 'l' ? triple_target::type_literal : triple_target::type_blank_node),
                               graph_buf, graph_length);

            ++ntriples;

            // output size so far
            if (super::total_read() > mibs_read)
            {
                clock_t current_time(clock());
                float  elapsed(float(current_time - start_time) / CLOCKS_PER_SEC);
                std::cerr << "#read " << super::total_read()/MiB << " MiB (" << (super::total_read()/MiB)/elapsed << " MiB/s).\t" << ntriples << " triples (" << ntriples/elapsed << " triples/s). " << std::endl;
                mibs_read += mibs_read_step;
            }

            // stop if we're only reading so far
            if (super::total_read() >= max_read)
            {
                break;
            }

            // continue to next triple
            int     punc(super::read());

            if (punc != ',')
            {
                if (punc == ';')
                {
                    if (!super::expect("\nINSERT INTO `triple_all` VALUES "))
                    {
                        if (strcmp(super::last(), "\nUNLOCK TABLES;\n/*!40000 ALTER TA"))
                        {
                            std::cerr << "expected insert statement at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                        }
                        break;
                    }
                }
                else
                {
                    std::cerr << "expected ',' at " << super::total_read() << ", found '" << super::last() << "'" << std::endl;
                    break;
                }
            }
        }
    }
};

#endif // _SQL_READER_H_
